from .circuits import QFT  # noqa: F401
from .circuits import DraperAdder  # noqa: F401
from .circuits import bell_pair  # noqa: F401
from .circuits import coin_toss  # noqa: F401
from .circuits import ghz  # noqa: F401
from .circuits import multi_adder  # noqa: F401
from .circuits import random_circuit  # noqa: F401



# Copyright 2023 Bluequbit Inc.
#
# All Rights Reserved
#
# NOTICE: All information contained herein is, and remains the property of Bluequbit Inc.
# and its suppliers, if any. The intellectual and technical concepts contained herein are
# proprietary to Bluequbit Inc. and its suppliers and may be covered by U.S. and Foreign Patents,
# patents in process, and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is strictly forbidden unless
# prior written permission is obtained from Bluequbit Inc.
#
# This file is part of a proprietary software product and is not to be copied, modified,
# redistributed, or used in any way without the express written permission of Bluequbit Inc.


import pennylane as qml
from pennylane import numpy as np
from pennylane_lightning.lightning_gpu.lightning_gpu import _gpu_dtype
from tqdm.notebook import trange

complex_dtype = np.complex64
real_dtype = np.float32


class AdamOptim:
    def __init__(self, eta=0.03, beta1=0.9, beta2=0.999, epsilon=1e-8):
        super().__init__()
        self.m, self.v = 0, 0
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.eta = eta
        self.t = 1

    def step(self, grad):
        ## dw, db are from current minibatch
        ## momentum beta 1
        self.m = self.beta1 * self.m + (1 - self.beta1) * grad
        ## rms beta 2
        self.v = self.beta2 * self.v + (1 - self.beta2) * (grad**2)
        ## bias correction
        m_corr = self.m / (1 - self.beta1**self.t)
        v_corr = self.v / (1 - self.beta2**self.t)
        self.t += 1
        ## update weights and biases
        return self.eta * (m_corr / (np.sqrt(v_corr) + self.epsilon))


def _get_circuit_info(circuit):
    # todo: assert circuit is qnode
    params_dummy = np.zeros(1000000, requires_grad=True)
    specs = qml.specs(circuit)(params_dummy)
    num_params = specs[
        "num_trainable_params"
    ]  # one way to get number of params from a circuit
    num_wires = specs["resources"].num_wires
    return num_wires, num_params


def load_data(
    target_probs: np.ndarray,
    circuit: qml.QNode,
    seed: int = 0,
    loss_type: str = "kl-divergence",  # noqa: ARG001
    num_epochs: int = 1000,
    params_init=None,
    eta=0.03,
) -> dict:
    r"""Run QCBM training to find an approximation to the target_distribution"""
    num_wires, num_params = _get_circuit_info(circuit)

    def adjoint_kl(params):
        dev_bra = qml.device("lightning.gpu", wires=num_wires, c_dtype=complex_dtype)
        dev_ket = qml.device("lightning.gpu", wires=num_wires, c_dtype=complex_dtype)
        dev_gpu = qml.device("lightning.gpu", wires=num_wires, c_dtype=complex_dtype)
        # tic = time.time()
        ket = circuit(params)
        ops = circuit.qtape.operations
        loss = np.sum(target_probs * np.log(target_probs / (np.abs(ket) ** 2)))
        # print("circuit_eval", time.time() - tic)
        bra = target_probs / ket.conj()
        bra_norm = np.linalg.norm(bra)
        bra /= bra_norm
        dev_bra.syncH2D(bra)
        # This line is important to convert NumPy array to PL tensor, so that
        # syncH2D is fast.
        ket = np.asarray(ket)
        dev_ket.syncH2D(ket)
        grads = []
        for op in reversed(ops):
            adj_op = qml.adjoint(op)
            dev_ket.apply([adj_op])
            if op.num_params != 0:
                du = qml.operation.operation_derivative(op)
                # Is this necessary for performance? It may change the result
                # du.requires_grad = False
                _qu = qml.QubitUnitary(du, op.wires)
                use_async = False
                dev_gpu._gpu_state.DeviceToDevice(  # noqa: SLF001
                    dev_ket._gpu_state,  # noqa: SLF001
                    use_async,
                )
                dev_gpu.apply([_qu])
                dm = (
                    -2
                    * bra_norm
                    * dev_gpu._gpu_state.dotWithBraReal(  # noqa: SLF001
                        dev_bra._gpu_state  # noqa: SLF001
                    )
                )
                # A more reliable way to reset _gpu_state
                # TODO raise a bug report to Pennylane
                dev_gpu._gpu_state = _gpu_dtype(complex_dtype)(  # noqa: SLF001
                    dev_gpu._num_local_wires  # noqa: SLF001
                )
                grads.append(dm)
            dev_bra.apply([adj_op])
        return grads[::-1], loss, ket

    np.random.seed(seed)
    best_x, best_loss = None, np.inf
    eps = 0.02
    if params_init is None:
        x = np.random.rand(num_params, requires_grad=True).astype(real_dtype)
    else:
        x = eps * np.random.rand(num_params, requires_grad=True).astype(real_dtype)
        x[: len(params_init)] = params_init

    print(f"starting training with qubits: {num_wires} params: {num_params}")
    print(qml.specs(circuit)(x)["resources"])
    #     print(qml.draw(circuit_gpu)(x))
    opt = AdamOptim(eta=eta)
    for epoch in trange(num_epochs):
        grad, loss, pred_state = adjoint_kl(x)
        grad = np.array(grad).astype(real_dtype)
        if loss < best_loss:
            best_loss = loss
            if best_loss < 5:
                l1_loss = np.sum(np.abs(target_probs - np.abs(pred_state) ** 2))
                print(f"NEW* best_loss: {best_loss}: L1_loss: {l1_loss}")
            best_x = x.copy()
        x -= opt.step(grad)
        print(f"epoch: {epoch} \t loss: {loss}", flush=True)
    return best_x, best_loss



from __future__ import annotations

from typing import TYPE_CHECKING

from qiskit.quantum_info import Statevector
from qiskit.result import Result
from qiskit.result.models import ExperimentResult, ExperimentResultData

import bluequbit

if TYPE_CHECKING:
    from qiskit import QuantumCircuit


class BlueQubitQiskitJob:
    def __init__(self, job_id: str, backend: BlueQubitBackend, jobs):
        super().__init__()
        self.job_id = job_id
        self.backend = backend
        jobs = jobs if isinstance(jobs, list) else [jobs]
        self.jobs = jobs

    def result(self):
        experiment_results: list[ExperimentResult] = []

        for job in self.jobs:
            finished = job
            if self.backend.device != "local":
                finished = self.backend.bq.wait(job.job_id)

            if job.shots in [0, None]:
                data = ExperimentResultData(
                    statevector=Statevector(finished.get_statevector()).data,
                )
            else:
                data = ExperimentResultData(
                    counts=finished.get_counts(),
                    statevector=(
                        Statevector(finished.get_statevector()).data
                        if job.num_qubits <= 15
                        else None
                    ),
                )
            experiment_results.append(
                ExperimentResult(
                    shots=job.shots,
                    success=True,
                    status=job.run_status,
                    data=data,
                )
            )
        first_job = self.jobs[0]
        return Result(
            backend_name=f"bluequbit_{first_job.device}",
            backend_version=bluequbit.__version__,
            job_id=first_job.job_id,
            qobj_id=0,
            # TODO this is too restricted
            success=first_job.run_status == "COMPLETED",
            results=experiment_results,
            status=first_job.run_status,
        )


class BlueQubitBackend:
    def __init__(self, bq=None, device="cpu"):
        super().__init__()
        if bq is None:
            bq = bluequbit.init()
        self.bq = bq
        self.device = device

    def run(
        self,
        run_input: QuantumCircuit | list[QuantumCircuit],
    ):
        jobs = self.bq.run(run_input, device=self.device, asynchronous=True)
        return BlueQubitQiskitJob(None, self, jobs)


class BlueQubitProvider:
    """A Qiskit provider for accessing BlueQubit backend.

    :param api_token: API token of the user. If ``None``, the token will be looked
                      in the environment variable BLUEQUBIT_API_TOKEN.
    """

    def __init__(self, api_token: str | None = None):
        super().__init__()
        self.bq = bluequbit.init(api_token=api_token)

    def get_backend(self, device: str = "cpu"):
        """
        :param device: device for which to run the circuit. Can be one of
                       ``"cpu"`` | ``"gpu"`` | ``"quantum"`` | ``"local"``
        """
        return BlueQubitBackend(bq=self.bq, device=device)



# Mypy; for the `|` operator purpose
# Remove this __future__ import once the oldest supported Python is 3.10
from __future__ import annotations

import contextlib
import logging

from . import exceptions, job_metadata_constants
from .bluequbit_client import BQClient
from .estimate_result import EstimateResult
from .job_result import JobResult

with contextlib.suppress(ImportError):
    from .qiskit_provider import BlueQubitProvider
from .version import __version__

__all__ = [
    "exceptions",
    "job_metadata_constants",
    "BQClient",
    "EstimateResult",
    "JobResult",
    "__version__",
    "BlueQubitProvider",
]

formatter = logging.Formatter(fmt="BQ-PYTHON-SDK - %(levelname)s - %(message)s")
# formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(module)s - %(message)s')

handler = logging.StreamHandler()
handler.setFormatter(formatter)

logger = logging.getLogger("bluequbit-python-sdk")
logger.setLevel(logging.INFO)
logger.addHandler(handler)


def init(api_token: str | None = None) -> BQClient:
    """Returns :class:`BQClient` instance for managing jobs on BlueQubit platform.

    :param api_token: API token of the user. If ``None``, the token will be looked
                      in the environment variable BLUEQUBIT_API_TOKEN.
    """
    return BQClient(api_token)



JOB_NO_RESULT_TERMINAL_STATES = {
    "FAILED_VALIDATION",
    "CANCELED",
    "TERMINATED",
    "NOT_ENOUGH_FUNDS",
    "JOBS_LIMIT_EXCEEDED",
}
JOB_RESULTS_READY_STATES = {"COMPLETED"}
JOB_TERMINAL_STATES = JOB_RESULTS_READY_STATES.union(JOB_NO_RESULT_TERMINAL_STATES)
JOB_NON_TERMINAL_STATES = {"PENDING", "QUEUED", "RUNNING"}
JOB_STATES = JOB_NON_TERMINAL_STATES.union(JOB_TERMINAL_STATES)

DEVICE_TYPES = {
    "cpu",
    "gpu",
    "quantum",
    "tensor-network",
    "local",
    "pennylane.cpu",
    "pennylane.gpu",
    "pennylane.cpu.adjoint",
    "pennylane.gpu.adjoint",
}

MAXIMUM_NUMBER_OF_BATCH_JOBS = 500
MAX_QUBITS_WITH_STATEVEC = 16



# Mypy; for the `|` operator purpose
# Remove this __future__ import once the oldest supported Python is 3.10
from __future__ import annotations

import contextlib
import logging
import os
import time
from typing import TYPE_CHECKING, Any, List, Tuple, Union

from . import job_metadata_constants
from .api import jobs
from .backend_connection import BackendConnection
from .check_version import check_version
from .circuit_serialization import isa_qiskit_circuit
from .computation_local import run_circuit_cirq, run_circuit_qiskit
from .estimate_result import EstimateResult
from .exceptions import (
    BQBatchJobsLimitExceededError,
    BQJobCouldNotCancelError,
    BQJobInvalidDeviceTypeError,
    BQJobNotCompleteError,
    BQSDKUsageError,
)
from .job_result import JobResult
from .version import __version__

if TYPE_CHECKING:
    import datetime

# TODO this requires imports of actual quantum libraries for proper type
# checking.
CircuitT = Any

PauliSumT = Union[List[Tuple[str, float]], str]

logger = logging.getLogger("bluequbit-python-sdk")


_SHOTS_LIMIT_NON_QPU = 131072


def _run_circuits_local(circuits, shots, backend_connection):
    created_on = time.time()

    def _format_result(r, num_qubits):
        def to_str_time(x):
            return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(x))

        r["created_on"] = to_str_time(created_on)
        r["run_start"] = to_str_time(r["run_start"])
        r["run_end"] = to_str_time(r["run_end"])
        # Rename counts to top_100k
        r["top_100k"] = r["counts"]
        del r["counts"]
        r["has_statevector"] = "state_vector" in r
        r["num_qubits"] = num_qubits
        r["run_status"] = "COMPLETED"
        return JobResult(r, backend_connection)

    first_circuit = circuits[0] if isinstance(circuits, list) else circuits
    try:
        from deqart_internal.circuit_info import get_num_qubits
        from deqart_internal.computation import run_circuit_object

        try:
            import cuquantum  # noqa: F401

            device_type = "gpu"
        except ImportError:
            # Not in a cuQuantum Appliance
            device_type = "cpu"
        print(f"Using cuQuantum cuStateVec with device type {device_type}")

        def _run_and_format_result(circuit):
            r = run_circuit_object(circuit, device_type, shots=shots)
            num_qubits = get_num_qubits(circuit)
            return _format_result(r, num_qubits)

        if isinstance(circuits, list):
            return [_run_and_format_result(c) for c in circuits]
        return _run_and_format_result(circuits)
    except (ImportError, ModuleNotFoundError):
        if isa_qiskit_circuit(first_circuit):
            fn = run_circuit_qiskit
        elif str(type(first_circuit)) == "<class 'cirq.circuits.circuit.Circuit'>":
            fn = run_circuit_cirq
        else:
            error_msg = f"Circuit type not yet supported for {first_circuit}"
            raise BQSDKUsageError(error_msg) from None

        if isinstance(circuits, list):
            return [_format_result(fn(c, shots=shots), len(c.qubits)) for c in circuits]
        return _format_result(fn(circuits, shots=shots), len(circuits.qubits))


class BQClient:
    """Client for managing jobs on BlueQubit platform.

    :param api_token: API token of the user. If ``None``, the token will be looked
                      in the environment variable BLUEQUBIT_API_TOKEN.
    """

    def __init__(self, api_token: str | None = None):
        super().__init__()
        if os.environ.get("BLUEQUBIT_TESTING") is None:
            with contextlib.suppress(Exception):
                check_version(__version__)

        self._backend_connection = BackendConnection(api_token)

    def name(self):
        return "BlueQubit"

    def validate_device(self, device):
        if not isinstance(device, str):
            raise BQJobInvalidDeviceTypeError(device)
        converted_device = device.lower()
        if converted_device not in job_metadata_constants.DEVICE_TYPES:
            raise BQJobInvalidDeviceTypeError(device)
        return converted_device

    @staticmethod
    def validate_batch(batch):
        if (
            isinstance(batch, list)
            and len(batch) > job_metadata_constants.MAXIMUM_NUMBER_OF_BATCH_JOBS
        ):
            raise BQBatchJobsLimitExceededError(len(batch))

    def estimate(
        self, circuits: CircuitT | list[CircuitT], device: str = "cpu"
    ) -> EstimateResult | list[EstimateResult]:
        """Estimate job runtime

        :param circuits: quantum circuit or circuits
        :type circuits: Cirq, Qiskit, list
        :param device: device for which to estimate the circuit. Can be one of
                       ``"cpu"`` | ``"gpu"`` | ``"quantum"``
        :return: result or results estimate metadata
        """
        device = self.validate_device(device)
        self.validate_batch(circuits)
        response = jobs.submit_jobs(
            self._backend_connection, circuits, device, estimate_only=True
        )
        if isinstance(circuits, list):
            return [EstimateResult(data) for data in response]
        return EstimateResult(response)

    def run(
        self,
        circuits: CircuitT | list[CircuitT],
        device: str = "cpu",
        asynchronous: bool = False,
        job_name: str | None = None,
        shots: int | None = None,
        pauli_sum: PauliSumT | list[PauliSumT] | None = None,
    ) -> JobResult | list[JobResult]:
        """Submit a job to run on BlueQubit platform

        :param circuits: quantum circuit or list of circuits
        :type circuits: Cirq, Qiskit, list
        :param device: device on which to run the circuit. Can be one of
                       ``"cpu"`` | ``"gpu"`` | ``"quantum"`` | ``"local"``
        :param asynchronous: if set to ``False``, wait for job completion before
                             returning. If set to ``True``, return immediately
        :param job_name: customizable job name
        :param shots: number of shots to run. If device is quantum and shots is None then
                      it is set to 1000. For non quantum devices, if None, full
                      probability distribution will be returned. For non quantum
                      devices it is limited to 131072
        :param pauli_sum: The Pauli sum or a list of Pauli sum which
                          expectation value is the computation result
        :return: job or jobs metadata
        """
        device = self.validate_device(device)
        self.validate_batch(circuits)
        if device == "quantum" and shots is None:
            shots = 1000
        elif device != "quantum" and shots is not None and shots > _SHOTS_LIMIT_NON_QPU:
            logger.warning(
                "Number of shots is set to %s, because of limit.",
                _SHOTS_LIMIT_NON_QPU,
            )
        elif device == "local":
            return _run_circuits_local(circuits, shots, self._backend_connection)
        response = jobs.submit_jobs(
            self._backend_connection,
            circuits,
            device,
            job_name,
            shots=shots,
            asynchronous=asynchronous,
            pauli_sum=pauli_sum,
        )
        if isinstance(circuits, list):
            logger.info(
                "Submitted %s jobs. Batch ID %s", len(response), response[0]["batch_id"]
            )

            def add_circuit_to_all(job_results, circuits):
                for job_result, circuit in zip(job_results, circuits):
                    job_result.circuit = circuit

            job_results = [
                JobResult(data, self._backend_connection) for data in response
            ]
            if not asynchronous:
                if self._check_all_in_terminal_states(job_results):
                    add_circuit_to_all(job_results, circuits)
                    return job_results

                job_results = self.wait(job_results)
                add_circuit_to_all(job_results, circuits)
                return job_results
                # if job_results[0].batch_id is not None:
                #     return self.wait(batch_id=job_results[0].batch_id)
                # else:
            add_circuit_to_all(job_results, circuits)
            return job_results
        submitted_job = JobResult(response, self._backend_connection)
        if (
            submitted_job.run_status
            in job_metadata_constants.JOB_NO_RESULT_TERMINAL_STATES
        ):
            raise BQJobNotCompleteError(
                submitted_job.job_id,
                submitted_job.run_status,
                submitted_job.error_message,
            )
        logger.info("Submitted: %s", submitted_job)
        if (
            not asynchronous
            and submitted_job.run_status
            not in job_metadata_constants.JOB_TERMINAL_STATES
        ):
            jr = self.wait(submitted_job.job_id)
            jr.circuit = circuits
            return jr
        submitted_job.circuit = circuits
        return submitted_job

    @staticmethod
    def _check_all_in_terminal_states(job_results):
        if not isinstance(job_results, list):
            return job_results.run_status in job_metadata_constants.JOB_TERMINAL_STATES
        return all(
            job_result.run_status in job_metadata_constants.JOB_TERMINAL_STATES
            for job_result in job_results
        )

    def wait(
        self, job_ids: str | list[str] | JobResult | list[JobResult]
    ) -> JobResult | list[JobResult]:
        """Wait for job completion

        :param job_ids: job IDs that can be found as property of :class:`JobResult` metadata
                        of :func:`~run` method, or `JobResult` instances from which job IDs
                        will be extracted
        :return: job metadata
        """
        self.validate_batch(job_ids)
        while True:
            job_results = self._get(job_ids, need_qc_unprocessed=False)
            if self._check_all_in_terminal_states(job_results):
                if (
                    not isinstance(job_ids, list)
                    and job_results.run_status
                    in job_metadata_constants.JOB_NO_RESULT_TERMINAL_STATES
                ):
                    raise BQJobNotCompleteError(
                        job_ids, job_results.run_status, job_results.error_message
                    )
                return job_results
            time.sleep(1.0)

    def get(
        self,
        job_ids: str | list[str] | JobResult | list[JobResult],
    ) -> JobResult | list[JobResult]:
        """Get current metadata of jobs

        :param job_ids: job IDs that can be found as property of :class:`JobResult` metadata
                        of :func:`~run` method
        :return: jobs metadata
        """
        return self._get(job_ids)

    def _get(
        self,
        job_ids: str | list[str] | JobResult | list[JobResult],
        need_qc_unprocessed=True,
    ) -> JobResult | list[JobResult]:
        self.validate_batch(job_ids)
        job_ids_list = job_ids if isinstance(job_ids, list) else [job_ids]
        if isinstance(job_ids_list[0], JobResult):
            job_ids_list = [jr.job_id for jr in job_ids_list]
        job_results = jobs.search_jobs(
            self._backend_connection,
            job_ids=job_ids_list,
            need_qc_unprocessed=need_qc_unprocessed,
        )
        job_results = [
            JobResult(r, self._backend_connection) for r in job_results["data"]
        ]
        if isinstance(job_ids, list):
            return job_results
        return job_results[0]

    def cancel(
        self, job_ids: str | list[str] | JobResult | list[JobResult]
    ) -> JobResult | list[JobResult]:
        """Submit jobs cancel request

        :param job_ids: job IDs that can be found as property of :class:`JobResult` metadata
                        of :func:`run` method
        :return: job or jobs metadata
        """
        self.validate_batch(job_ids)
        if isinstance(job_ids, JobResult):
            job_ids = job_ids.job_id
        elif isinstance(job_ids, list) and isinstance(job_ids[0], JobResult):
            job_ids = [jr.job_id for jr in job_ids]
        responses = jobs.cancel_jobs(self._backend_connection, job_ids)
        if isinstance(job_ids, list):
            for response in responses:
                if response["ret"] == "FAILED":
                    logger.warning(response["error_message"])
        try:
            self.wait(job_ids)
        except BQJobNotCompleteError as e:
            if not e.run_status == "CANCELED":
                raise BQJobCouldNotCancelError(
                    e.job_id, e.run_status, e.error_message
                ) from None
        return self.get(job_ids)

    def search(
        self,
        run_status: str | None = None,
        created_later_than: str | datetime.datetime | None = None,
        batch_id: str | None = None,
    ) -> list[JobResult]:
        """Search jobs

        :param run_status: if not ``None``, run status of jobs to filter.
                           Can be one of ``"FAILED_VALIDATION"`` | ``"PENDING"`` |
                           ``"QUEUED"`` | ``"RUNNING"`` | ``"TERMINATED"`` | ``"CANCELED"`` |
                           ``"NOT_ENOUGH_FUNDS"`` | ``"COMPLETED"``

        :param created_later_than: if not ``None``, filter by latest job creation datetime.
                                   Please add timezone for clarity, otherwise UTC
                                   will be assumed

        :param batch_id: if not ``None``, filter by batch ID

        :return: metadata of jobs
        """
        job_results = jobs.search_jobs(
            self._backend_connection, run_status, created_later_than, batch_id=batch_id
        )
        return [JobResult(r, self._backend_connection) for r in job_results["data"]]



import json
import logging
import os
from pathlib import Path

from .exceptions import AUTH_ERROR_MESSAGE, BQAPIError, BQUnauthorizedAccessError
from .http_utils import request_retriable
from .version import __version__

logger = logging.getLogger("bluequbit-python-sdk")


class BackendConnection:
    def __init__(self, api_token=None):
        super().__init__()
        config_dir = Path.home() / ".config" / "bluequbit"
        config_location = config_dir / "config.json"

        main_endpoint_from_config_file = None
        ssl_verify_from_config_file = None
        token_from_env_variable = os.environ.get("BLUEQUBIT_API_TOKEN")

        if config_location.is_file():
            with config_location.open(encoding="utf-8") as f:
                config = json.load(f)
            main_endpoint_from_config_file = config.get("main_endpoint")
            ssl_verify_from_config_file = config.get("ssl_verify")
        self._authenticated = None
        if api_token is None:
            if token_from_env_variable is None:
                logger.warning(AUTH_ERROR_MESSAGE)
                self._authenticated = False
            api_token = token_from_env_variable

        api_config = {
            "token": api_token,
            "main_endpoint": (
                "https://app.bluequbit.io/api/v1"
                if main_endpoint_from_config_file is None
                else main_endpoint_from_config_file
            ),
            "ssl_verify": (
                True
                if ssl_verify_from_config_file is None
                else ssl_verify_from_config_file
            ),
        }

        self._token = api_token

        self._default_headers = {
            "Authorization": f"SDK {self._token}",
            "Connection": "close",
            "User-Agent": f"BlueQubit SDK {__version__}",
        }

        self._main_endpoint = "https://app.bluequbit.io/api/v1"
        if self._main_endpoint != api_config["main_endpoint"]:
            self._main_endpoint = api_config["main_endpoint"]
            logger.warning("Using custom endpoint %s", self._main_endpoint)
        self._verify = True
        if "ssl_verify" in api_config:
            self._verify = api_config["ssl_verify"]
        self._session = None
        if self._authenticated is not False:
            try:
                response = self.send_request(
                    req_type="GET",
                    path="/jobs",
                    params={"limit": 1},
                )
                if response.ok:
                    self._authenticated = True
                else:
                    self._authenticated = False
                    raise BQAPIError(
                        response.status_code,
                        f"Couldn't reach BlueQubit {response.text}.",
                    )
            except BQUnauthorizedAccessError:
                self._authenticated = False
                logger.warning(AUTH_ERROR_MESSAGE)
                return

    def send_request(
        self,
        req_type,
        path,
        params=None,
        data=None,
        json_req=None,
        headers=None,
    ):
        url = self._main_endpoint + path

        if params is not None:
            for key, value in params.items():
                if isinstance(value, str):
                    params[key] = value.replace("\\", "\\\\")
                if isinstance(value, list):
                    params[key] = ",".join(value)

        if headers is None:
            headers_to_send = self._default_headers
        else:
            headers_to_send = dict(self._default_headers, **headers)

        return request_retriable(
            method=req_type,
            url=url,
            data=data,
            json=json_req,
            params=params,
            headers=headers_to_send,
        )
